==> transformers4rec-0.1.1.tar.gz: transformers4rec-0.1.1/tests/tf/model/test_head.py <==
#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest

tf = pytest.importorskip("tensorflow")
tr = pytest.importorskip("transformers4rec.tf")
test_utils = pytest.importorskip("transformers4rec.tf.utils.testing_utils")


@pytest.mark.parametrize("prediction_task", [tr.BinaryClassificationTask, tr.RegressionTask])
def test_simple_heads(tf_tabular_features, tf_tabular_data, prediction_task):
    targets = {"target": tf.cast(tf.random.uniform((100,), maxval=2, dtype=tf.int32), tf.float32)}

    body = tr.SequentialBlock([tf_tabular_features, tr.MLPBlock([64])])
    task = prediction_task("target")
    head = task.to_head(body, tf_tabular_features)

    test_utils.assert_loss_and_metrics_are_valid(head, tf_tabular_data, targets)


@pytest.mark.parametrize("prediction_task", [tr.BinaryClassificationTask, tr.RegressionTask])
def test_serialization_simple_heads(tf_tabular_features, tf_tabular_data, prediction_task):
    targets = {"target": tf.cast(tf.random.uniform((100,), maxval=2, dtype=tf.int32), tf.float32)}

    body = tr.SequentialBlock([tf_tabular_features, tr.MLPBlock([64])])
    task = prediction_task("target")
    head = task.to_head(body, tf_tabular_features)

    copy_head = test_utils.assert_serialization(head)
    test_utils.assert_loss_and_metrics_are_valid(copy_head, tf_tabular_data, targets)


@pytest.mark.parametrize("task", [tr.BinaryClassificationTask, tr.RegressionTask])
@pytest.mark.parametrize("task_block", [None, tr.MLPBlock([32])])
@pytest.mark.parametrize("summary", ["last", "first", "mean", "cls_index"])
def test_simple_heads_on_sequence(
    tf_yoochoose_tabular_sequence_features, tf_yoochoose_like, task, task_block, summary
):
    inputs = tf_yoochoose_tabular_sequence_features
    targets = {"target": tf.cast(tf.random.uniform((100,), maxval=2, dtype=tf.int32), tf.float32)}

    body = tr.SequentialBlock([inputs, tr.MLPBlock([64])])
    head = task("target", task_block=task_block, summary_type=summary).to_head(body, inputs)

    test_utils.assert_loss_and_metrics_are_valid(head, tf_yoochoose_like, targets)


@pytest.mark.parametrize(
    "task_blocks",
    [
        None,
        tr.MLPBlock([32]),
        dict(classification=tr.MLPBlock([16]), regression=tr.MLPBlock([20])),
        dict(binary_classification_task=tr.MLPBlock([16]), regression_task=tr.MLPBlock([20])),
        {
            "classification/binary_classification_task": tr.MLPBlock([16]),
            "regression/regression_task": tr.MLPBlock([20]),
        },
    ],
)
def test_head_with_multiple_tasks(tf_tabular_features, tf_tabular_data, task_blocks):
    targets = {
        "classification": tf.cast(tf.random.uniform((100,), maxval=2, dtype=tf.int32), tf.float32),
        "regression": tf.cast(tf.random.uniform((100,), maxval=2, dtype=tf.int32), tf.float32),
    }

    body = tr.SequentialBlock([tf_tabular_features, tr.MLPBlock([64])])
    tasks = [
        tr.BinaryClassificationTask("classification"),
        tr.RegressionTask("regression"),
    ]
    head = tr.Head(body, tasks, task_blocks=task_blocks)
    model = tr.Model(head)
    model.compile(optimizer="adam")

    step = model.train_step((tf_tabular_data, targets))

    # assert 0 <= step["loss"] <= 1 # test failing with loss greater than 1
    assert step["loss"] >= 0
    assert len(step) == 8
    if task_blocks:
        task_blocks = list(head.task_blocks.values())
        assert task_blocks[0] != task_blocks[1]


def test_item_prediction_head_shape(tf_yoochoose_tabular_sequence_features, tf_yoochoose_like):

    body = tr.SequentialBlock(
        [tf_yoochoose_tabular_sequence_features, tr.MLPBlock([64])],
    )

    task = tr.NextItemPredictionTask(weight_tying=True)
    head = task.to_head(body)

    outputs = head(body(tf_yoochoose_like))

    assert outputs.shape[-1] == 51997


# Test loss and metrics outputs
@pytest.mark.parametrize("weight_tying", [True, False])
def test_item_prediction_loss_and_metrics(
    tf_yoochoose_tabular_sequence_features, tf_yoochoose_like, weight_tying
):
    body = tr.SequentialBlock(
        [tf_yoochoose_tabular_sequence_features, tr.MLPBlock([64])],
    )

    task = tr.NextItemPredictionTask(weight_tying=True)
    head = task.to_head(body)

    body_outputs = body(tf_yoochoose_like)

    outputs = head(body_outputs)

    loss = head.compute_loss(body_outputs=body_outputs, targets=None)

    metrics = head.metric_results()
    assert len(metrics) == 6
    default_metric = [
        "ndcg_at_10",
        "ndcg_at_20",
        "avg_precision_at_10",
        "avg_precision_at_20",
        "recall_at_10",
        "recall_at_20",
    ]
    assert set(default_metric).issubset(set(metrics.keys()))
    assert outputs.shape[-1] == 51997
    assert loss != 0


def test_item_prediction_head_with_wrong_body(tf_tabular_features, tf_tabular_data):
    with pytest.raises(ValueError) as excinfo:
        body = tr.SequentialBlock([tf_tabular_features, tr.MLPBlock([64])])
        task = tr.NextItemPredictionTask(weight_tying=True)
        head = task.to_head(body)
        _ = head(body(tf_tabular_data))

        assert "NextItemPredictionTask needs a 3-dim vector as input, found:" in str(excinfo.value)

==> transformers4rec-0.1.2.tar.gz: transformers4rec-0.1.2/tests/tf/model/test_head.py <==
#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest

tf = pytest.importorskip("tensorflow")
tr = pytest.importorskip("transformers4rec.tf")
test_utils = pytest.importorskip("transformers4rec.tf.utils.testing_utils")


@pytest.mark.parametrize("prediction_task", [tr.BinaryClassificationTask, tr.RegressionTask])
def test_simple_heads(tf_tabular_features, tf_tabular_data, prediction_task):
    targets = {"target": tf.cast(tf.random.uniform((100,), maxval=2, dtype=tf.int32), tf.float32)}

    body = tr.SequentialBlock([tf_tabular_features, tr.MLPBlock([64])])
    task = prediction_task("target")
    head = task.to_head(body, tf_tabular_features)

    test_utils.assert_loss_and_metrics_are_valid(head, tf_tabular_data, targets)


@pytest.mark.parametrize("prediction_task", [tr.BinaryClassificationTask, tr.RegressionTask])
def test_serialization_simple_heads(tf_tabular_features, tf_tabular_data, prediction_task):
    targets = {"target": tf.cast(tf.random.uniform((100,), maxval=2, dtype=tf.int32), tf.float32)}

    body = tr.SequentialBlock([tf_tabular_features, tr.MLPBlock([64])])
    task = prediction_task("target")
    head = task.to_head(body, tf_tabular_features)

    copy_head = test_utils.assert_serialization(head)
    test_utils.assert_loss_and_metrics_are_valid(copy_head, tf_tabular_data, targets)


def test_serialization_item_prediction_head(tf_yoochoose_like, yoochoose_schema):
    input_module = tr.TabularSequenceFeatures.from_schema(
        yoochoose_schema,
        max_sequence_length=20,
        continuous_projection=64,
        d_output=64,
    )
    body = tr.SequentialBlock([input_module, tr.MLPBlock([64])])
    task = tr.NextItemPredictionTask(weight_tying=True, metrics=[])

    head = task.to_head(body, input_module)

    copy_head = test_utils.assert_serialization(head)
    targets = tf_yoochoose_like["item_id/list"]

    loss = copy_head.compute_loss(tf_yoochoose_like, targets, call_body=True)
    assert loss is not None


@test_utils.mark_run_eagerly_modes
def test_item_prediction_yoochoose_model(yoochoose_schema, tf_yoochoose_like, run_eagerly):
    input_module = tr.TabularSequenceFeatures.from_schema(
        yoochoose_schema,
        max_sequence_length=20,
        continuous_projection=64,
        d_output=64,
    )
    body = tr.SequentialBlock([input_module, tr.MLPBlock([64])])
    task = tr.NextItemPredictionTask(weight_tying=True, metrics=[])

    model = task.to_model(body, input_module)
    model.compile(optimizer="adam", run_eagerly=run_eagerly)

    outputs = model(tf_yoochoose_like)
    assert outputs.shape[-1] == 51997


@pytest.mark.parametrize("task", [tr.BinaryClassificationTask, tr.RegressionTask])
@pytest.mark.parametrize("task_block", [None, tr.MLPBlock([32])])
@pytest.mark.parametrize("summary", ["last", "first", "mean", "cls_index"])
def test_simple_heads_on_sequence(
    tf_yoochoose_tabular_sequence_features, tf_yoochoose_like, task, task_block, summary
):
    inputs = tf_yoochoose_tabular_sequence_features
    targets = {"target": tf.cast(tf.random.uniform((100,), maxval=2, dtype=tf.int32), tf.float32)}

    body = tr.SequentialBlock([inputs, tr.MLPBlock([64])])
    head = task("target", task_block=task_block, summary_type=summary).to_head(body, inputs)

    test_utils.assert_loss_and_metrics_are_valid(head, tf_yoochoose_like, targets)


@pytest.mark.parametrize(
    "task_blocks",
    [
        None,
        tr.MLPBlock([32]),
        dict(classification=tr.MLPBlock([16]), regression=tr.MLPBlock([20])),
        dict(binary_classification_task=tr.MLPBlock([16]), regression_task=tr.MLPBlock([20])),
        {
            "classification/binary_classification_task": tr.MLPBlock([16]),
            "regression/regression_task": tr.MLPBlock([20]),
        },
    ],
)
def test_head_with_multiple_tasks(tf_tabular_features, tf_tabular_data, task_blocks):
    targets = {
        "classification": tf.cast(tf.random.uniform((100,), maxval=2, dtype=tf.int32), tf.float32),
        "regression": tf.cast(tf.random.uniform((100,), maxval=2, dtype=tf.int32), tf.float32),
    }

    body = tr.SequentialBlock([tf_tabular_features, tr.MLPBlock([64])])
    tasks = [
        tr.BinaryClassificationTask("classification"),
        tr.RegressionTask("regression"),
    ]
    head = tr.Head(body, tasks, task_blocks=task_blocks)
    model = tr.Model(head)
    model.compile(optimizer="adam")

    step = model.train_step((tf_tabular_data, targets))

    # assert 0 <= step["loss"] <= 1 # test failing with loss greater than 1
    assert step["loss"] >= 0
    assert len(step) == 8
    if task_blocks:
        task_blocks = list(head.task_blocks.values())
        assert task_blocks[0] != task_blocks[1]


def test_item_prediction_head_shape(tf_yoochoose_tabular_sequence_features, tf_yoochoose_like):

    body = tr.SequentialBlock(
        [tf_yoochoose_tabular_sequence_features, tr.MLPBlock([64])],
    )

    task = tr.NextItemPredictionTask(weight_tying=True)
    head = task.to_head(body)

    outputs = head(body(tf_yoochoose_like))

    assert outputs.shape[-1] == 51997


# Test loss and metrics outputs
@pytest.mark.parametrize("weight_tying", [True, False])
def test_item_prediction_loss_and_metrics(
    tf_yoochoose_tabular_sequence_features, tf_yoochoose_like, weight_tying
):
    body = tr.SequentialBlock(
        [tf_yoochoose_tabular_sequence_features, tr.MLPBlock([64])],
    )

    task = tr.NextItemPredictionTask(weight_tying=True)
    head = task.to_head(body)

    body_outputs = body(tf_yoochoose_like)

    loss = head.compute_loss(body_outputs=body_outputs, targets=None)

    metrics = head.metric_results()
    assert len(metrics) == 6
    default_metric = [
        "ndcg_at_10",
        "ndcg_at_20",
        "avg_precision_at_10",
        "avg_precision_at_20",
        "recall_at_10",
        "recall_at_20",
    ]
    assert set(default_metric).issubset(set(metrics.keys()))
    assert loss != 0


def test_item_prediction_head_with_wrong_body(tf_tabular_features, tf_tabular_data):
    with pytest.raises(ValueError) as excinfo:
        body = tr.SequentialBlock([tf_tabular_features, tr.MLPBlock([64])])
        task = tr.NextItemPredictionTask(weight_tying=True)
        head = task.to_head(body)
        _ = head(body(tf_tabular_data))

        assert "NextItemPredictionTask needs a 3-dim vector as input, found:" in str(excinfo.value)
